# TiCS: disabled # samba mock

dom_sid = ""
descriptor = "SECURITY_DESCRIPTOR"

SECINFO_OWNER = 1<<0
SECINFO_GROUP = 1<<1
SECINFO_DACL = 1<<2

SEC_STD_READ_CONTROL = 1<<0
SEC_ADS_LIST = 1<<1
SEC_ADS_READ_PROP = 1<<2
